#
# Ruby/Migemo - a library for Japanese incremental search.
#
# Copyright (C) 2001 Satoru Takabayashi <satoru@namazu.org>
#     All rights reserved.
#     This is free software with ABSOLUTELY NO WARRANTY.
#
# You can redistribute it and/or modify it under the terms of 
# the GNU General Public License version 2.
#
# NOTE: Ruby/Migemo can work only with EUC_JP encoding. ($KCODE="e")
#

require 'migemo-dict'
require 'romkan'
require 'jcode'

class String
  # Hiragana to Katakana
  def to_katakana
    self.gsub(//, '\\1').tr('-', '-')
  end

  def quotemeta
    self.gsub(/([^ \w])/, '\\\\\\1')
  end

  def first
    /^(\\.|.)/ =~ self
    $1
  end

  def last
    /(\\.|.)$/ =~ self
    $1
  end

  def rest
    /^(\\.|.)(.*)/ =~ self
    $2
  end

  HANZEN_TAB = {
    " " => "", "!" => "", '"' => "", "#" => "", 
    "\$" => "", "%" => "", "&" => "", "'" => "",
    "(" => "", ")" => "", "*" => "", "+" => "",
    "," => "", "-" => "", "." => "", "/" => "",
    "0" => "", "1" => "", "2" => "", "3" => "",
    "4" => "", "5" => "", "6" => "", "7" => "",
    "8" => "", "9" => "", ":" => "", ";" => "",
    "<" => "", "=" => "", ">" => "", "?" => "",
    '@' => "", "A" => "", "B" => "", "C" => "",
    "D" => "", "E" => "", "F" => "", "G" => "",
    "H" => "", "I" => "", "J" => "", "K" => "",
    "L" => "", "M" => "", "N" => "", "O" => "",
    "P" => "", "Q" => "", "R" => "", "S" => "",
    "T" => "", "U" => "", "V" => "", "W" => "",
    "X" => "", "Y" => "", "Z" => "", "[" => "", 
    "\\" => "", "]" => "", "^" => "", "_" => "",
    "`" => "", "a" => "", "b" => "", "c" => "",
    "d" => "", "e" => "", "f" => "", "g" => "",
    "h" => "", "i" => "", "j" => "", "k" => "",
    "l" => "", "m" => "", "n" => "", "o" => "",
    "p" => "", "q" => "", "r" => "", "s" => "",
    "t" => "", "u" => "", "v" => "", "w" => "",
    "x" => "", "y" => "", "z" => "", "{" => "",
    "|" => "", "}" => "", "~" => ""} #'

  HANZEN_RE = Regexp.new(HANZEN_TAB.keys.sort.map {|x| x.quotemeta}.join('|'))

  def to_fullwidth
    self.gsub(HANZEN_RE) {|s| HANZEN_TAB[s]}
  end
end

class RegexAlternation < Array
  def sort
    self.clone.replace(super)
  end

  def uniq
    self.clone.replace(super)
  end

  def map
    self.clone.replace(super {|x| yield(x)})
  end

  def delete_if
    self.clone.replace(super {|x| yield(x)})
  end

  def select
    self.clone.replace(super {|x| yield(x)})
  end
end

class RegexConcatnation < Array
  def map
    self.clone.replace(super {|x| yield(x)})
  end
end

class RegexCharClass < Array
end

class RegexCompiler
  def initialize
    @regex = RegexAlternation.new
  end
  attr_reader :regex

  private
  # ["", "ư", "ž", "", "ܸ"] => ["" "" ""]
  # (|ư|ž||ܸ) => (||)
  def optimize1 (regex)
    prefixpat = nil
    regex.sort.select do |word|
      if prefixpat && prefixpat.match(word) then
	false # excluded
      else
	prefixpat = Regexp.new(word.quotemeta)
	true # included
      end
    end
  end

  # (||)
  # => (((||)))
  def optimize2 (regex)
    tmpregex = regex.sort.clone # I wish Array#cdr were available...
    optimized = RegexAlternation.new
    until tmpregex.empty?
      head = tmpregex.shift
      initial = head.first
      friends = RegexAlternation.new
      while item = tmpregex.first
	if initial == item.first
	  friends.push(item.rest)
	  tmpregex.shift
	else
	  break
	end
      end
      if friends.empty?
	optimized.push head
      else
	concat = RegexConcatnation.new
	concat.push(initial)
	friends.unshift(head.rest) 
	concat.push(optimize2(friends))
	optimized.push(concat)
      end
    end
    return optimized
  end

  # (||||)
  # => []
  def optimize3 (regex)
    charclass = RegexCharClass.new
    if regex.instance_of?(RegexAlternation)
      regex.delete_if do |x|
	if x.instance_of?(String) && x =~ /^.$/ then
	  charclass.push(x)
	  true
	end
      end
    end

    if charclass.length == 1
      regex.unshift charclass.first
    elsif charclass.length > 1
      regex.unshift charclass
    end

    regex.map do |x|
      if x.instance_of?(RegexAlternation) || x.instance_of?(RegexConcatnation)
	optimize3(x)
      else
	x
      end
    end
  end

  public
  def push (item)
    @regex.push(item)
  end

  def uniq
    @regex.uniq
  end

  def optimize (level)
    @regex = optimize1(@regex) if level >= 1
    @regex = optimize2(@regex) if level >= 2
    @regex = optimize3(@regex) if level >= 3
  end
end

class RegexRenderer
  def initialize (regex, insertion)
    raise if regex == nil
    @regex = regex
    @meta = Hash.new
    @meta['bar'] = '|'
    @meta['lparen'] = '('
    @meta['rparen'] = ')'
    @insertion = insertion
  end

  private
  def render_alternation (regex)
    if regex.length == 0
      raise
    elsif regex.length == 1
      render0(regex[0])
    else
      @meta['lparen'] + 
	regex.map {|x| render0(x) }.join(@meta['bar']) + 
	@meta['rparen']
    end
  end

  def render_concatnation (regex)
    regex.map {|x| render0(x) }.join(@insertion)
  end

  def escape_string (string)
    string.quotemeta
  end

  def escape_charclass (string)
    string.gsub(/\\/, '\\\\\\')
  end

  def render_charclass (regex)
    if regex.delete("-")
      regex.push("-")  # move "-" to the end of Array.
    end
    if regex.delete("]")
      regex.unshift("]")  # move "]" to the beginning of Array.
    end
    escape_charclass("[" + regex.join + "]")
  end

  def insert (string)
    if @insertion != ""
      tmp = string.gsub(/(\\.|.)/, "\\1#{@insertion}")
      tmp = tmp.sub(/#{@insertion.quotemeta}$/, "")
    else
      string
    end
  end

  def render_string (regex)
    insert(escape_string(regex))
  end

  def render0 (x)
    if x.instance_of?(RegexAlternation)
      render_alternation(x)
    elsif x.instance_of?(RegexConcatnation)
      render_concatnation(x)
    elsif x.instance_of?(RegexCharClass)
      render_charclass(x)
    elsif x.instance_of?(String)
      render_string(x)
    else
      raise "unexpected type: #{x} of #{x.class}"
    end
  end

  public
  def render
    @regex.map do |x|
      render0(x)
    end.join @meta['bar']
  end

  def join_regexes (string, regexes)
    ([string] + regexes).join @meta['bar']
  end
end

class RegexEmacsRenderer < RegexRenderer
  def initialize (regex, insertion)
    super(regex, insertion)
    @meta['bar'] = '\\|'
    @meta['lparen'] = '\\('
    @meta['rparen'] = '\\)'
  end

  def escape_string (string)
    str = string.quotemeta
    str.gsub!(/\\\(/, "(")
    str.gsub!(/\\\)/, ")")
    str.gsub!(/\\\|/, "|")
    str.gsub!(/\\\</, "<")
    str.gsub!(/\\\>/, ">")
    str.gsub!(/\\\=/, "=")
    str.gsub!(/\\\'/, "'")
    str.gsub!(/\\\`/, "`")
    str.gsub!(/\\\{/, "{")
    str
  end

  def escape_charclass (string)
    string
  end
end

module RegexRendererFactory
  def new (regex, type, insertion)
    case type
    when "emacs"
      RegexEmacsRenderer.new(regex, insertion)
    when "perl"
      RegexRenderer.new(regex, insertion)
    when "ruby"
      RegexRenderer.new(regex, insertion)
    else
      raise "Unknown type: #{regex}"
    end
  end
  module_function :new
end

class Migemo
  VERSION = '0.3'
  def initialize (dict, pattern)
    @type = "ruby"
    @pattern = pattern
    @insertion = ""
    @optimization = 3
    @static_dict = dict
    @dict_cache = nil
    @user_dict = nil
    @regex_dict = nil
  end
  attr_accessor :optimization
  attr_accessor :type
  attr_accessor :insertion
  attr_accessor :dict_cache
  attr_accessor :user_dict
  attr_accessor :regex_dict

  private
  # `do'   => ()
  # `d'    => (     )
  # `sh'   => (    )
  # `don'  => (ɤ ɤ ɤ ɤ ɤ ɤ ɤ)  # special case 1
  # `nodd' => (Τ)                                # special case 2
  # `doc'  => (ɤ ɤ)                           # special case 3
  # `dox'  => (ɤ ɤ ɤ ɤ)                 # special case 4
  def expand_kanas
    kana = @pattern.to_kana
    /^(.*)(.)$/ =~ kana ;
    head = $1;
    last = $2;

    cand = Array.new;
    if last.consonant?
      if /^(.*)(.)$/ =~ head && $2.consonant?
	head2 = $1;
	beforelast = $2;
	if last == $beforelast # special case 2
	  cand.push head2 + ""
	else
	  cand += (beforelast + last).expand_consonant.map do |x|
	    head2 + x.to_kana
	  end
	end
      else
	deriv = last.expand_consonant
	deriv.push "xtsu";
	if last == "c" # special case 3
	  deriv.push "chi";
	elsif last == "x" # special case 4
	  deriv.push "xya", "xyu", "xyo", "xwa"
	end
	cand += deriv.map do |x| head + x.to_kana end
      end
    elsif last == "" # speacial case 1
      cand.push kana;
      cand += ("n".expand_consonant + [""]).map do |x|
	head + x.to_kana
      end
    else
      cand.push kana
    end
    return cand.sort
  end

  # `meshi' => ( ӹ   ݤ ܲ γ  Ӳ)
  def expand_words0 (dict, pattern)
    raise if pattern == nil
    words = Array.new
    dict.lookup(pattern) do |item|
      words += item.values
    end

    # Handle special cases.
    if /^(.*)(.)$/ =~ pattern
      head = $1;
      last = $2;
      if last == "z" 
	dict.lookup(head + "j") do |item|
	  words += item.values
	end
      elsif last == "t" 
	dict.lookup(head + "ch") do |item|
	  words += item.values
	end
      elsif last == "y" && /^(.*)s$/ =~ head   # sy
	dict.lookup($1 + "sh") do |item|
	  words += item.values
	end
      elsif last == "y" && /^(.*)t$/ =~ head   # ty
	dict.lookup($1 + "ch") do |item|
	  words += item.values
	end
      end
    end
    return words
  end

  def expand_words (dict, pattern)
    words = expand_words0(dict, pattern)
    hepburn = pattern.to_hepburn
    if hepburn != nil && hepburn != pattern
      words += expand_words0(dict, hepburn)
    end
    words
  end

  def lookup_cache
    @dict_cache.lookup(@pattern)
  end

  def lookup0
    compiler = RegexCompiler.new
    compiler.push(@pattern)
    compiler.push(@pattern.to_fullwidth)
    expand_kanas.each do |x| 
      compiler.push(x)
      compiler.push(x.to_katakana)
    end
    expand_words(@static_dict, @pattern).each do |x| compiler.push(x) end
    compiler.uniq
    compiler.optimize(@optimization) if @optimization
    compiler.regex
  end

  def lookup_user_dict
    compiler = RegexCompiler.new
    expand_words(@user_dict, @pattern).each do |x| compiler.push(x) end
    compiler.uniq
    compiler.optimize(@optimization) if @optimization
    compiler.regex
  end

  def lookup_regex_dict
    regexes = []
    @regex_dict.lookup(@pattern) do |item|
      regexes += item.values
    end
    regexes
  end

  public
  def lookup
    result = if @dict_cache
	       lookup_cache || lookup0
	     else
	       lookup0
	     end
    result += lookup_user_dict if @user_dict
    result
  end

  def regex_tree
    lookup
  end

  def regex
    regex = lookup
    renderer = RegexRendererFactory.new(regex, @type, @insertion)
    string = renderer.render
    string = renderer.join_regexes(string, lookup_regex_dict) if @regex_dict
    string
  end
end

